#include <QsLog.h>
#include <QTimer>
#include <qmath.h>
#include "dbus_service_monitor.h"
#include "defines.h"
#include "battery_life.h"
#include "multi.h"
#include "settings.h"
#include "system_calc.h"

class Guard
{
public:
	Guard(bool *busy):
		mBusy(busy)
	{
		Q_ASSERT(!*mBusy);
		*mBusy = true;
	}

	~Guard()
	{
		*mBusy = false;
	}

private:
	bool *mBusy;
};

static const double AbsorptionLevel = 85;
static const double FloatLevel = 95;
static const QString SystemCalcService = "com.victronenergy.system";
static const QString SocPath = "/Dc/Battery/Soc";

BatteryLife::BatteryLife(SystemCalc *systemCalc, Multi *multi,
						 Settings *settings, Clock *clock, QObject *parent):
	QObject(parent),
	mSystemCalc(systemCalc),
	mMulti(multi),
	mSettings(settings),
	mClock(clock == 0 ? new DefaultClock() : clock),
	mUpdateBusy(false)
{
	Q_ASSERT(systemCalc != 0);
	Q_ASSERT(multi != 0);
	Q_ASSERT(settings != 0);
	connect(mSystemCalc, SIGNAL(socChanged()), this, SLOT(update()));
	connect(mMulti, SIGNAL(isSustainActiveChanged()), this, SLOT(onSustainActiveChanged()));
	connect(mSettings, SIGNAL(socLimitChanged()), this, SLOT(onSocLimitChanged()));
	connect(mSettings, SIGNAL(stateChanged()), this, SLOT(onStateChanged()));
	connect(mSettings, SIGNAL(minSocLimitChanged()), this, SLOT(onMinimumSocChanged()));

	QTimer *chargeTimer = new QTimer(this);
	chargeTimer->setInterval(SocInterval);
	connect(chargeTimer, SIGNAL(timeout()), this, SLOT(onChargeTimer()));

	update();
	chargeTimer->start();
}

void BatteryLife::onChargeTimer()
{
	QDateTime now = mClock->now();
	if (QTime(0, 0, 0).msecsTo(now.time()) < SocInterval)
		mSettings->setFlags(0);
	if (mSettings->state() == BatteryLifeStateDischarged ||
		mSettings->state() == BatteryLifeStateSustain) {
		QDateTime dischargeTime = mSettings->dischargedTime();
		if (dischargeTime.isValid())
		{
			if (dischargeTime.msecsTo(now) >= ForceChargeInterval) {
				adjustSocLimit(SocSwitchIncrement);
				setState(BatteryLifeStateForceCharge);
			}
		} else {
			// This should only happen if the discharge time in the	settings is changed externally
			// when the state is 'discharged'.
			mSettings->setDischargedTime(mClock->now());
		}
	}
}

void BatteryLife::update()
{
	if (mUpdateBusy)
		return;
	Guard guard(&mUpdateBusy);
	double soc = mSystemCalc->soc();
	if (!qIsFinite(soc)) {
		QLOG_TRACE() << "[Battery life] No active battery or no valid SoC";
		return;
	}
	double socLimit = qBound(0.0, qMax(mSettings->socLimit(), mSettings->minSocLimit()), SocSwitchMax);
	BatteryLifeState state = mSettings->state();
	QLOG_TRACE() << "[Battery life]" << soc << socLimit << getStateName(state);
	bool adjustLimit = true;
	switch (mSettings->state()) {
	case BatteryLifeStateDisabled:
		break;
	case BatteryLifeStateRestart:
		adjustLimit = false;
		// Fall through
	case BatteryLifeStateDefault:
		// Switch off check
		if (mMulti->isSustainActive() || soc <= socLimit) {
			onDischarged(adjustLimit);
		} else if (soc >= FloatLevel) {
			onFloat(adjustLimit);
		} else if (soc >= AbsorptionLevel) {
			onAbsorption(adjustLimit);
		} else {
			setState(BatteryLifeStateDefault);
		}
		break;
	case BatteryLifeStateSustain:
	case BatteryLifeStateDischarged:
		// Switch on check
		if (!mMulti->isSustainActive() && soc > socLimit + SocSwitchOffset)
			setState(BatteryLifeStateDefault);
		break;
	case BatteryLifeStateForceCharge:
		if (!mMulti->isSustainActive() && soc > socLimit) {
			mSettings->setDischargedTime(mClock->now());
			setState(BatteryLifeStateDischarged);
		}
		break;
	case BatteryLifeStateAbsorption:
		if (mMulti->isSustainActive())
			onDischarged(adjustLimit);
		else if (soc >= FloatLevel)
			onFloat(adjustLimit);
		else if (soc < AbsorptionLevel - SocSwitchOffset)
			setState(BatteryLifeStateDefault);
		break;
	case BatteryLifeStateFloat:
		if (mMulti->isSustainActive())
			onDischarged(adjustLimit);
		else if (soc < FloatLevel - SocSwitchOffset)
			setState(BatteryLifeStateAbsorption);
		break;
	default:
		setState(BatteryLifeStateDefault);
		break;
	}
}

void BatteryLife::onStateChanged()
{
	QLOG_INFO() << "[Battery life] State changed to" << getStateName(mSettings->state());
	update();
}

void BatteryLife::onSocLimitChanged()
{
	QLOG_INFO() << "[Battery life] SoC limit changed to" << mSettings->socLimit();
	update();
}

void BatteryLife::onMinimumSocChanged()
{
	QLOG_INFO() << "[Battery life] Minimum SoC changed to" << mSettings->minSocLimit();
	update();
}

void BatteryLife::onSustainActiveChanged()
{
	QLOG_INFO() << "[Battery life] Sustain enabled" << mMulti->isSustainActive();
	update();
}

void BatteryLife::onDischarged(bool adjustLimit)
{
	if (adjustLimit) {
		if ((mSettings->flags() & BatteryLifeFlagDischarged) == 0) {
			mSettings->setFlags(mSettings->flags() | BatteryLifeFlagDischarged);
			adjustSocLimit(SocSwitchIncrement);
		}
		mSettings->setDischargedTime(mClock->now());
	}
	setState(mMulti->isSustainActive() ? BatteryLifeStateSustain : BatteryLifeStateDischarged);
}

void BatteryLife::onAbsorption(bool adjustLimit)
{
	if (adjustLimit && (mSettings->flags() & BatteryLifeFlagAbsorption) == 0) {
		mSettings->setFlags(mSettings->flags() | BatteryLifeFlagAbsorption);
		adjustSocLimit(-SocSwitchIncrement);
	}
	setState(BatteryLifeStateAbsorption);
}

void BatteryLife::onFloat(bool adjustLimit)
{
	if (adjustLimit && (mSettings->flags() & BatteryLifeFlagFloat) == 0) {
		mSettings->setFlags(mSettings->flags() | BatteryLifeFlagFloat);
		double offset = -SocSwitchIncrement;
		// If the charger synchronizes its SoC it may skip the SoC range we consider to be
		// absorption, so we add the usual offset twice.
		if (mSettings->state() != BatteryLifeStateAbsorption)
			offset *= 2;
		adjustSocLimit(offset);
	}
	setState(BatteryLifeStateFloat);
}

void BatteryLife::adjustSocLimit(double delta)
{
	double socLimit = qMax(mSettings->minSocLimit(), mSettings->socLimit()) + delta;
	socLimit = qBound(0.0, socLimit, SocSwitchMax);
	mSettings->setSocLimit(socLimit);
}

void BatteryLife::setState(BatteryLifeState state)
{
	mSettings->setState(state);
}

const char *BatteryLife::getStateName(BatteryLifeState state)
{
	static const char *StateNames[] = { "Disabled", "Restart", "Default", "Absorption", "Float",
										"Discharged", "ForceCharge", "Sustain" };
	static const int StateNameCount = static_cast<int>(sizeof(StateNames)/sizeof(StateNames[0]));
	if (state < 0 || state >= StateNameCount)
		return "Unknown";
	return StateNames[state];
}
