#include <qmath.h>
#include <QsLog.h>
#include <QTimer>
#include "battery_info.h"
#include "control_loop.h"
#include "defines.h"
#include "grid_meter.h"
#include "multi.h"
#include "multi_phase_data.h"
#include "pvinverter.h"
#include "pv_power_limiter.h"
#include "settings.h"
#include "system_calc.h"

ControlLoop::ControlLoop(SystemCalc *systemCalc, Multi *multi, GridMeter *gridMeter,
						 Settings *settings, QObject *parent):
	QObject(parent),
	mSystemCalc(systemCalc),
	mGridMeter(gridMeter),
	mMulti(multi),
	mSettings(settings),
	mBatteryInfo(0),
	mPVPowerLimiter(0)
{
	Q_ASSERT(systemCalc != 0);
	Q_ASSERT(multi != 0);
	Q_ASSERT(settings != 0);
}

void ControlLoop::setBatteryInfo(BatteryInfo *c)
{
	mBatteryInfo = c;
}

void ControlLoop::setPVPowerLimiter(PVPowerLimiter *l)
{
	mPVPowerLimiter = l;
}

double ControlLoop::getGridPower(Phase phase) const
{
	if (mGridMeter == 0)
		return mMulti->getPhaseData(phase)->acPowerIn();
	return mGridMeter->getPower(phase);
}

void ControlLoop::adjustSetpoint(double gridPower, Phase targetPhase, MultiPhaseData *target,
								 double setpoint)
{
	if (!qIsFinite(target->acPowerSetPoint()))
		return;
	if (mMulti->firmwareVersion() < 0x405) {
		// Older vebus firmware versions had a faster control loop on the AC-In power setpoint.
		// In order to prevent overshoot we reduce the difference between the new setpoint and the
		// current AC-In power.
		// Old code was: setpoint = 0.8 * change + acPowerIn. with change = setpoint - acPowerIn
		setpoint = 0.2 * target->acPowerIn() + 0.8 * setpoint;
	}

	if (qIsFinite(setpoint)) {
		setpoint = qBound(MinMultiPower, setpoint, MaxMultiPower);

		// Ugly workaround: the value of pMultiNew must always be sent over the
		// D-Bus, even when it does not change, because the multi will reset its
		// power setpoint if no value has been set during the last 10 seconds.
		setpoint = qRound(setpoint);
		if (qRound(target->acPowerSetPoint()) == setpoint)
			setpoint -= 1;
	}

	// If feedback is disabled and the multi is 'on', it will start inverting
	// when the grid drops ways (emergency power). If we set the multi to
	// 'charge only', it will not start inverting so the system on AC-In and
	// AC-Out will lose power.
	// mMulti->setMode(feedbackDisabled ? MultiChargerOnly : MultiOn);
	QLOG_TRACE() << QString("L%1").arg(targetPhase)
				 << gridPower
				 << '\t' << target->acPowerIn()
				 << '\t' << target->acPowerSetPoint()
				 << '\t' << setpoint
				 << '\t' << mMulti->isChargeDisabled()
				 << '\t' << mMulti->isFeedbackDisabled();
	if (qIsFinite(setpoint))
		target->setAcPowerSetPoint(setpoint);
}

double ControlLoop::adjustTotalSetpoint(double setpoint)
{
	bool feedbackDisabled = true;

	switch (mSettings->state()) {
	case BatteryLifeStateForceCharge:
	{
		double dcCurrent = mMulti->dcCurrent();
		double dcVoltage = mMulti->dcVoltage();
		if (qIsFinite(dcCurrent) && qIsFinite(dcCurrent)) {
			double dcPower = dcCurrent * dcVoltage;
			double acPower = 0;
			double acPhaseOut = 0;
			for (int i=0; i<3; ++i) {
				Phase phase = static_cast<Phase>(PhaseL1 + i);
				MultiPhaseData *mpd = mMulti->getPhaseData(phase);
				double acPowerIn = mpd->acPowerIn();
				if (qIsFinite(acPowerIn))
					acPower += acPowerIn;
				double acPowerOut = mpd->acPowerOut();
				if (qIsFinite(acPowerOut)) {
					acPhaseOut = acPowerOut;
					acPower -= acPowerOut;
				}
			}
			double minAcPower = dcVoltage * ForceChargeCurrent;
			if (dcPower > 0.1 && acPower > 0)
				minAcPower *= qBound(1.0, acPower / dcPower, 2.0);
			minAcPower += acPhaseOut;
			setpoint = qMax(minAcPower, setpoint);
		}
		break;
	}
	case BatteryLifeStateSustain:
	case BatteryLifeStateDischarged:
		break;
	default:
	{
		double systemSoc = mSystemCalc->soc();
		double minSocLimit = qBound(0.0, mSettings->minSocLimit(), 100.0);
		feedbackDisabled = !mBatteryInfo->canDischarge() || systemSoc <= minSocLimit;
		break;
	}
	}

	if (mBatteryInfo != 0)
		setpoint = mBatteryInfo->applyLimits(setpoint);

	// It seems that enabling the ChargeDisabled flag disables both charge and discharge. So
	// we're only setting the flags when we are not discharging.
	bool chargeDisabled = !mBatteryInfo->canCharge() && (
		feedbackDisabled || (qIsFinite(setpoint) && setpoint > 30));

	mMulti->setIsChargeDisabled(chargeDisabled);
	mMulti->setIsFeedbackDisabled(feedbackDisabled);

	return setpoint;
}

bool ControlLoop::hasSetpoint(Phase phase) const
{
	return mMulti->getPhaseData(phase)->isSetPointAvailable();
}

double ControlLoop::pvPowerSetpointOffset() const
{
	return mPVPowerLimiter == 0 ? 0 : mPVPowerLimiter->pvPowerSetpointOffset();
}
