#include <QsLog.h>
#include <velib/qt/ve_qitem.hpp>
#include "battery_info.h"
#include "battery_life.h"
#include "dbus_service_monitor.h"
#include "hub4_control.h"
#include "hub4_control_bridge.h"
#include "multi.h"
#include "multi_bridge.h"
#include "multi_phase_data.h"
#include "phase_compensation_control.h"
#include "pv_power_limiter.h"
#include "settings.h"
#include "settings_bridge.h"
#include "settings_converter.h"
#include "single_phase_control.h"
#include "split_phase_control.h"
#include "vbus_item_battery.h"
#include "vbus_item_gridmeter.h"
#include "vbus_item_system_calc.h"
#include "vbus_item_pvinverter.h"

Hub4Control::Hub4Control(QObject *parent):
	QObject(parent),
	mServiceMonitor(new DbusServiceMonitor(this)),
	mSettings(new Settings(this)),
	mGridMeter(0),
	mMulti(new Multi(this)),
	mSystemCalc(new VBusItemSystemCalc(this)),
	mBatteryLife(0),
	mControlLoop(0),
	mBatteryInfo(new BatteryInfo(mMulti, mSettings)),
	mPVPowerLimiter(new PVPowerLimiter(mSystemCalc, mMulti, mSettings, this)),
	mHub4ControlBridge(0),
	mTimeZone(VeQItems::getRoot()->itemGetOrCreate("sub/com.victronenergy.settings/Settings/System/TimeZone")),
	mHub4Mode(VeQItems::getRoot()->itemGetOrCreate("sub/com.victronenergy.settings/Settings/CGwacs/Hub4Mode"))
{
	qRegisterMetaType<BatteryLifeState>();
	qRegisterMetaType<MultiMode>();

	mHub4ControlBridge = new Hub4ControlBridge(mBatteryInfo, mSettings, this);

	connect(mTimeZone, SIGNAL(valueChanged(VeQItem *, QVariant)), this, SLOT(onTimeZoneChanged()));
	mTimeZone->getValue();

	connect(mHub4Mode, SIGNAL(valueChanged(VeQItem *, QVariant)), this, SLOT(updateControlLoop()));
	mHub4Mode->getValue();

	new SettingsBridge(mSettings, this);

	SettingsConverter *c = new SettingsConverter(VeQItems::getRoot()->itemGetOrCreate("sub"), this);
	c->start();
	connect(c, SIGNAL(finished()), c, SLOT(deleteLater()));

	for (int i=0; i<3; ++i) {
		Phase phase = static_cast<Phase>(PhaseL1 + i);
		connect(mMulti->getPhaseData(phase), SIGNAL(isSetPointAvailableChanged()),
				this, SLOT(updateControlLoop()));
	}
	connect(mServiceMonitor, SIGNAL(serviceAdded(VeQItem *)),
			this, SLOT(onServiceAdded(VeQItem *)));
	connect(mServiceMonitor, SIGNAL(serviceRemoved(VeQItem *)),
			this, SLOT(onServiceRemoved(VeQItem *)));
	mServiceMonitor->start();
}

void Hub4Control::onTimeZoneChanged()
{
	QString timeZone = mTimeZone->getValue().toString();
	setenv("TZ", timeZone.toLatin1(), 1);
	QDateTime time = QDateTime::currentDateTime();
	QLOG_INFO() << "Changed time zone to:" << timeZone
				<< "current local time is:" << time.toString();
}

void Hub4Control::onSerialEvent(const char *description)
{
	QLOG_ERROR() << "Serial event:" << description
				 << "Application will shut down.";
	exit(1);
}

void Hub4Control::onServiceAdded(VeQItem *root)
{
	QString serviceName = root->id();
	if (serviceName.startsWith("com.victronenergy.vebus.")) {
		MultiBridge *bridge = mMulti->findChild<MultiBridge *>();
		if (bridge == 0) {
			QLOG_INFO() << "Multi found @" << serviceName;
			new MultiBridge(mMulti, root, mMulti);
		}
	} else if (serviceName.startsWith("com.victronenergy.battery.")) {
		QLOG_INFO() << "Battery found @" << serviceName;
		Battery *battery = new VbusItemBattery(root, this);
		mBatteryInfo->addBattery(battery);
	} else if (serviceName.startsWith("com.victronenergy.pvinverter.")) {
		QLOG_INFO() << "PV inverter found @" << serviceName;
		PVInverter *inverter = new VBusItemPVInverter(root, this);
		if (mPVPowerLimiter != 0)
			mPVPowerLimiter->addPVInverter(inverter);
	} else if (serviceName.startsWith("com.victronenergy.grid.")) {
		QLOG_INFO() << "Grid meter found @" << serviceName;
		if (mGridMeter == 0)
			setupGridMeter(root);
		if (!mGridMeterServices.contains(serviceName))
			mGridMeterServices.append(serviceName);
	}
}

void Hub4Control::onServiceRemoved(VeQItem *root)
{
	QString serviceName = root->id();
	if (serviceName.startsWith("com.victronenergy.vebus.")) {
		MultiBridge *bridge = mMulti->findChild<MultiBridge *>();
		if (bridge != 0 && bridge->service() == root) {
			QLOG_INFO() << "Multi @" << serviceName << "disappeared.";
			delete bridge;
			bridge = 0;
		}
	} else if (serviceName.startsWith("com.victronenergy.battery.")) {
		QLOG_INFO() << "Battery @" << serviceName << "disappeared.";
		foreach (VbusItemBattery *b, findChildren<VbusItemBattery *>()) {
			if (b->service() == root) {
				mBatteryInfo->removeBattery(b);
				delete b;
				break;
			}
		}
	} else if (serviceName.startsWith("com.victronenergy.pvinverter.")) {
		QLOG_INFO() << "PV inverter @" << serviceName << "disappeared.";
		foreach (VBusItemPVInverter *inverter, findChildren<VBusItemPVInverter *>()) {
			if (inverter->service() == root) {
				if (mPVPowerLimiter != 0)
					mPVPowerLimiter->removePVInverter(inverter);
				delete inverter;
			}
		}
	} else if (serviceName.startsWith("com.victronenergy.grid.")) {
		QLOG_INFO() << "Grid meter @" << serviceName << "disappeared.";
		mGridMeterServices.removeOne(serviceName);
		if (mGridMeter->service() == root) {
			delete mGridMeter;
			mGridMeter = 0;
			mPVPowerLimiter->setGridMeter(0);
			updateControlLoop();
			if (!mGridMeterServices.isEmpty()) {
				VeQItem *newRoot = VeQItems::getRoot()->itemGetOrCreate("sub/" + mGridMeterServices.first(), false);
				setupGridMeter(newRoot);
			}
		}
	}
}

void Hub4Control::updateControlLoop()
{
	delete mControlLoop;
	mControlLoop = 0;
	if ((!mSettings->runWithoutGridMeter() && mGridMeter == 0) || !mHub4Mode->getValue().isValid()) {
		QLOG_INFO() << "Control loop: no energy meter with service type 'grid'";
		delete mBatteryLife;
		mBatteryLife = 0;
		return;
	}
	if (mBatteryLife == 0)
		mBatteryLife = new BatteryLife(mSystemCalc, mMulti, mSettings, 0, this);
	Hub4Mode hub4Mode = static_cast<Hub4Mode>(mHub4Mode->getValue().toInt());
	if (mGridMeter == 0 || mGridMeter->isMultiPhase()) {
		switch (hub4Mode) {
		case Hub4SinglePhaseCompensation:
		{
			QList<Phase> setpointPhases = mMulti->getSetpointPhases();
			Phase phase = setpointPhases.isEmpty() ? PhaseL1 : setpointPhases.first();
			QLOG_INFO() << QString("Control loop: multi phase, phase L%1").arg(phase);
			mControlLoop = new SinglePhaseControl(mSystemCalc, mMulti, mGridMeter, mSettings, phase,
												  Hub4PhaseCompensation, this);
			break;
		}
		case Hub4PhaseCompensation:
			QLOG_INFO() << "Control loop: multi phase, phase compensation";
			mControlLoop = new PhaseCompensationControl(mSystemCalc, mMulti, mGridMeter, mSettings, this);
			break;
		case Hub4PhaseSplit:
			QLOG_INFO() << "Control loop: split phase control";
			mControlLoop = new SplitPhaseControl(mSystemCalc, mMulti, mGridMeter, mSettings, this);
			break;
		case Hub4PhaseL1:
		case Hub4Disabled:
		default:
			break;
		}
	} else {
		switch (hub4Mode) {
		case Hub4Disabled:
			break;
		default:
			QLOG_INFO() << "Control loop: single phase";
			mControlLoop = new SinglePhaseControl(mSystemCalc, mMulti, mGridMeter, mSettings,
												  PhaseL1, Hub4PhaseL1, this);
			break;
		}
	}
	if (mControlLoop == 0) {
		QLOG_INFO() << "Control loop: disabled";
	} else {
		mControlLoop->setBatteryInfo(mBatteryInfo);
		if (mPVPowerLimiter != 0)
			mControlLoop->setPVPowerLimiter(mPVPowerLimiter);
	}
}

void Hub4Control::setupGridMeter(VeQItem *root)
{
	VeQItem *settingsRoot = VeQItems::getRoot()->itemGetOrCreate("sub/com.victronenergy.settings", false);
	mGridMeter = new VBusItemGridMeter(root, settingsRoot, this);
	connect(mGridMeter, SIGNAL(isMultiPhaseChanged()), this, SLOT(updateControlLoop()));
	mPVPowerLimiter->setGridMeter(mGridMeter);
	updateControlLoop();
}
