#include <QsLog.h>
#include <QTimer>
#include "grid_meter.h"
#include "battery_info.h"
#include "multi.h"
#include "multi_phase_control.h"
#include "multi_phase_data.h"

MultiPhaseControl::MultiPhaseControl(SystemCalc *systemCalc, Multi *multi, GridMeter *gridMeter,
									 Settings *settings, QObject *parent):
	ControlLoop(systemCalc, multi, gridMeter, settings, parent),
	mTimer(new QTimer(this)),
	mMultiL1Update(false),
	mMultiL2Update(false),
	mMultiL3Update(false),
	mMultiTotalPowerUpdate(false),
	mMeterUpdate(false)
{
	Q_ASSERT(multi != 0);
	Q_ASSERT(settings != 0);
	mTimer->setInterval(6000);
	mTimer->start();
	connect(multi->getPhaseData(PhaseL1), SIGNAL(acPowerInChanged()),
			this, SLOT(onL1FromMulti()));
	connect(multi->getPhaseData(PhaseL2), SIGNAL(acPowerInChanged()),
			this, SLOT(onL2FromMulti()));
	connect(multi->getPhaseData(PhaseL3), SIGNAL(acPowerInChanged()),
			this, SLOT(onL3FromMulti()));
	connect(multi->getPhaseData(MultiPhase), SIGNAL(acPowerInChanged()),
			this, SLOT(onTotalPowerFromMulti()));
	if (gridMeter != 0) {
		connect(gridMeter, SIGNAL(powerChanged(Phase)),
				this, SLOT(onPowerFromMeter(Phase)));
	}
	connect(mTimer, SIGNAL(timeout()), this, SLOT(onTimer()));
}

void MultiPhaseControl::adjustSetpoints(const QVector<double> &setpoints)
{
	double setpointTotal = 0;
	for (int p=0; p<3; ++p) {
		if (qIsFinite(setpoints[p]))
			setpointTotal += setpoints[p];
	}
	/// @todo EV adjustTotalSetpoint has side effects
	double sp = adjustTotalSetpoint(setpointTotal);
	double f = (setpointTotal == 0) ? 1 : sp / setpointTotal;
	Multi *multi = this->multi();
	for (int p=0; p<3; ++p) {
		Phase phase = static_cast<Phase>(PhaseL1 + p);
		MultiPhaseData *mpd = multi->getPhaseData(phase);
		if (qIsFinite(setpoints[p]))
			adjustSetpoint(getGridPower(phase), phase, mpd, setpoints[p] * f);
	}
}

void MultiPhaseControl::onTimer()
{
	QLOG_DEBUG() << "Update timeout";
	mMultiL1Update = true;
	mMultiL2Update = true;
	mMultiL3Update = true;
	mMeterUpdate = true;
	checkStep();
}

void MultiPhaseControl::onL1FromMulti()
{
	mMultiL1Update = true;
	checkStep();
}

void MultiPhaseControl::onL2FromMulti()
{
	mMultiL2Update = true;
	checkStep();
}

void MultiPhaseControl::onL3FromMulti()
{
	mMultiL3Update = true;
	checkStep();
}

void MultiPhaseControl::onTotalPowerFromMulti()
{
	mMultiTotalPowerUpdate = true;
	checkStep();
}

void MultiPhaseControl::onPowerFromMeter(Phase phase)
{
	/// @todo EV Check each phase, just like Multi?
	// Total power is changed after phase specific data. So if total power changes, L1, L2, and L3
	// have fresh values as well.
	if (phase != MultiPhase)
		return;
	mMeterUpdate = true;
	checkStep();
}

void MultiPhaseControl::checkStep()
{
	if (!mMultiTotalPowerUpdate)
		return;
	if (!mMeterUpdate && acSensor() != 0)
		return;
	if (!mMultiL1Update && hasSetpoint(PhaseL1))
		return;
	if (!mMultiL2Update && hasSetpoint(PhaseL2))
		return;
	if (!mMultiL3Update && hasSetpoint(PhaseL3))
		return;
	performStep();
	mMeterUpdate = false;
	mMultiL1Update = false;
	mMultiL2Update = false;
	mMultiL3Update = false;
	mMultiTotalPowerUpdate = false;
	mTimer->start();
}
