#include <QTimer>
#include <QsLog.h>
#include <qmath.h>
#include "grid_meter.h"
#include "multi.h"
#include "multi_phase_data.h"
#include "pvinverter.h"
#include "pv_power_limiter.h"
#include "settings.h"
#include "system_calc.h"

const int MaxProbeCount = 20;

PVPowerLimiter::PVPowerLimiter(SystemCalc *systemCalc, Multi *multi, Settings *settings,
	QObject *parent):
	QObject(parent),
	mGridMeter(0),
	mSystemCalc(systemCalc),
	mMulti(multi),
	mSettings(settings),
	mPvInverterProbeTimer(0),
	mState(Probe),
	mPvLimit(0),
	mPrevError(0),
	mAvgError(0),
	mAvgErrorCount(0),
	mAvgPower(0),
	mAvgConsumption(0),
	mProbeCount(MaxProbeCount)
{
	QTimer *timer = new QTimer(this);
	timer->setInterval(2500);
	timer->start();
	connect(timer, SIGNAL(timeout()), this, SLOT(updateLimit()));
}

double PVPowerLimiter::pvPowerSetpointOffset() const
{
	if (!mSettings->preventFeedback())
		return 0;
	foreach (PVInverter *p, mPVInverters) {
		if (qIsFinite(p->powerLimit())) {
			return mState == Probe ? p->minPowerLimit() : 1.5 * p->powerLimitStepSize();
		}
	}
	return 0;
}

void PVPowerLimiter::addPVInverter(PVInverter *pvInverter)
{
	if (!mPVInverters.contains(pvInverter))
		mPVInverters.append(pvInverter);
}

void PVPowerLimiter::removePVInverter(PVInverter *pvInverter)
{
	mPVInverters.removeOne(pvInverter);
}

void PVPowerLimiter::setGridMeter(GridMeter *gridMeter)
{
	mGridMeter = gridMeter;
}

void PVPowerLimiter::updateLimit()
{
	/// @todo EV Support for multiple PV inverters.
	/// @todo EV If multiple PV inverters present, which power step size to use?
	PVInverter *pvInverter = 0;
	foreach (PVInverter *p, mPVInverters) {
		if (qIsFinite(p->powerLimit())) {
			pvInverter = p;
			break;
		}
	}

	if (pvInverter == 0) {
		setPvPowerLimitState(Probe);
		return;
	}

	if (!mSettings->preventFeedback() || !mMulti->isConnected()) {
		setPvPowerLimitState(Probe);
		pvInverter->setPowerLimit(pvInverter->maxPower());
		return;
	}

	double totalGridPower = mGridMeter == 0 ? mMulti->meanData()->acPowerIn() : mGridMeter->getTotalPower();
	double gridSetpoint = mSettings->acPowerSetPoint();
	double error = totalGridPower - gridSetpoint;

	double minPowerLimit = pvInverter->minPowerLimit();
	mAvgPower = 0.9 * mAvgPower + 0.1 * pvInverter->power();
	double consumption = mMulti->meanData()->acPowerOut() - mMulti->meanData()->acPowerIn() + error;
	if (qIsFinite(consumption))
		mAvgConsumption = 0.7 * mAvgConsumption + 0.3 * consumption;

	switch (mState) {
	case On:
	{
		double stepSize = pvInverter->powerLimitStepSize();
		mPvLimit += qBound(-stepSize, 0.1 * (error - mPrevError) + 0.4 * error, stepSize);
		mPvLimit = qMin(mAvgPower + stepSize, mPvLimit);
		mPrevError = error;
		pvInverter->setPowerLimit(qMax(minPowerLimit, qCeil(mPvLimit / stepSize) * stepSize));
		if (mPvLimit < minPowerLimit) {
			mAvgError = 0.9 * mAvgError + 0.1 * error;
			++mAvgErrorCount;
			if (mAvgErrorCount > 10 && mAvgError < 0)
				setPvPowerLimitState(Off);
		} else {
			mAvgErrorCount = 0;
			mAvgError = 0;
		}
		break;
	}
	case Off:
		if (mAvgConsumption > minPowerLimit)
			setPvPowerLimitState(Probe);
		else
			pvInverter->setPowerLimit(0);
		break;
	case Probe:
		if (mAvgConsumption <= minPowerLimit && mSystemCalc != 0 && mSystemCalc->soc() > 90) {
			setPvPowerLimitState(Off);
		} else {
			--mProbeCount;
			if (mProbeCount <= 0 || pvInverter->power() > 0) {
				mPvLimit = minPowerLimit;
				setPvPowerLimitState(On);
			} else {
				pvInverter->setPowerLimit(minPowerLimit);
			}
		}
		break;
	default:
		break;
	}
}

void PVPowerLimiter::onPvInverterProbeTimer()
{
	if (mState == Off) {
		if (mSystemCalc != 0 && mSystemCalc->soc() < 90)
			setPvPowerLimitState(Probe);
		else
			mPvInverterProbeTimer->start();
	}
}

void PVPowerLimiter::setPvPowerLimitState(PvPowerLimitState s)
{
	if (mState == s)
		return;
	QLOG_INFO() << "[PV power limiter] Power limit state:" << s;
	switch (s) {
	case On:
		mPvLimit = 0;
		mPrevError = 0;
		mAvgErrorCount = 0;
		mAvgError = 0;
		break;
	case Off:
		if (mPvInverterProbeTimer == 0) {
			mPvInverterProbeTimer = new QTimer(this);
			mPvInverterProbeTimer->setInterval(30 * 60 * 1000);
			mPvInverterProbeTimer->setSingleShot(true);
			connect(mPvInverterProbeTimer, SIGNAL(timeout()), this, SLOT(onPvInverterProbeTimer()));
		}
		mPvInverterProbeTimer->start();
		break;
	case Probe:
		mProbeCount = MaxProbeCount;
		break;
	}
	mState = s;
}
