#include <dbus_bridge.h>
#include <QDebug>
#include <velib/qt/ve_qitem.hpp>
#include "defines.h"
#include "settings_converter.h"

SettingsConverter::SettingsConverter(VeQItem *itemRoot, QObject *parent):
	QObject(parent),
	mSettingsRoot(itemRoot->itemGetOrCreate("com.victronenergy.settings", false)),
	mHub4Mode(mSettingsRoot->itemGetOrCreate("Settings/CGwacs/Hub4Mode")),
	mDeviceIds(0),
	mServiceType(0),
	mDeviceHub4Mode(0),
	mConversionNeeded(false),
	mFinished(false)
{
}

void SettingsConverter::start()
{
	onHub4ModeFound();
}

void SettingsConverter::onHub4ModeFound()
{
	switch (mHub4Mode->getState())
	{
	case VeQItem::Idle:
		mHub4Mode->getValue();
		// Fall through
	case VeQItem::Requested:
		connect(mHub4Mode, SIGNAL(stateChanged(VeQItem *, State)),
				this, SLOT(onHub4ModeFound()));
		break;
	case VeQItem::Synchronized:
		if (mConversionNeeded) {
			mDeviceIds = mSettingsRoot->itemGetOrCreate("Settings/CGwacs/DeviceIds");
			onDeviceIdsFound();
		} else {
			Q_ASSERT(mHub4Mode->getValue().isValid());
			setFinished();
		}
		break;
	case VeQItem::Offline:
	{
		// Create setting
		mConversionNeeded = true;
		DBusBridge b(mSettingsRoot, false);
		b.addSetting("/Settings/CGwacs/Hub4Mode", Hub4PhaseCompensation, 0, 0);
		break;
	}
	default:
		break;
	}
}

void SettingsConverter::onDeviceIdsFound()
{
	switch (mDeviceIds->getState())
	{
	case VeQItem::Idle:
		mDeviceIds->getValue();
		// Fall through
	case VeQItem::Requested:
		connect(mDeviceIds, SIGNAL(stateChanged(VeQItem *, State)),
				this, SLOT(onDeviceIdsFound()));
		break;
	case VeQItem::Synchronized:
		mDeviceIdsLeft = mDeviceIds->getValue().toString().split(',', QString::SkipEmptyParts);
		findNextDevice();
		break;
	case VeQItem::Offline:
		setFinished();
		break;
	default:
		break;
	}
}

void SettingsConverter::onServiceTypeFound()
{
	switch (mServiceType->getState())
	{
	case VeQItem::Idle:
		mServiceType->getValue();
		// Fall through
	case VeQItem::Requested:
		connect(mServiceType, SIGNAL(stateChanged(VeQItem *, State)),
				this, SLOT(onServiceTypeFound()));
		break;
	case VeQItem::Synchronized:
		if (mServiceType->getValue().toString() == "grid") {
			QString path = QString("Settings/CGwacs/Devices/D%1/Hub4Mode").arg(mDeviceIdsLeft.last());
			if (mDeviceHub4Mode != 0)
				mDeviceHub4Mode->disconnect(this);
			mDeviceHub4Mode = mSettingsRoot->itemGetOrCreate(path, true);
			onDeviceHub4ModeFound();
		} else {
			mDeviceIdsLeft.removeLast();
			findNextDevice();
		}
		break;
	case VeQItem::Offline:
		mDeviceIdsLeft.removeLast();
		findNextDevice();
		break;
	default:
		break;
	}
}

void SettingsConverter::onDeviceHub4ModeFound()
{
	switch (mDeviceHub4Mode->getState())
	{
	case VeQItem::Idle:
		mDeviceHub4Mode->getValue();
		// Fall through
	case VeQItem::Requested:
		connect(mDeviceHub4Mode, SIGNAL(stateChanged(VeQItem *, State)),
				this, SLOT(onDeviceHub4ModeFound()));
		break;
	case VeQItem::Synchronized:
	{
		Q_ASSERT(mDeviceHub4Mode->getValue().isValid());
		mConversionNeeded = false;
		int mode = mDeviceHub4Mode->getValue().toInt();
		// Conversion from dbus-cgwacs v1.0.8 to v1.0.9: Compensating over Phase 1 is no longer used
		// for multi phase systems.
		if (mode == Hub4PhaseL1)
			mode = Hub4PhaseCompensation;
		mHub4Mode->setValue(mode);
		setFinished();
		break;
	}
	case VeQItem::Offline:
		findNextDevice();
		break;
	default:
		break;
	}
}

void SettingsConverter::findNextDevice()
{
	if (mDeviceIdsLeft.isEmpty()) {
		setFinished();
	} else {
		QString id = mDeviceIdsLeft.last();
		QString path = QString("Settings/CGwacs/Devices/D%1/ServiceType").arg(id);
		if (mServiceType != 0)
			mServiceType->disconnect(this);
		mServiceType = mSettingsRoot->itemGetOrCreate(path, true);
		onServiceTypeFound();
	}
}

void SettingsConverter::setFinished()
{
	if (mFinished)
		return;
	mFinished = true;
	emit finished();
}
